/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee                                                **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "Executable.h"
#include "elg_error.h"

#include <iostream>
   using std::cerr;
   using std::cout;
   using std::endl;

typedef map<string, Module*>::const_iterator   MCI;
typedef map<string, Function*>::const_iterator FCI;
typedef map<string, Module*>::iterator         MI;
typedef map<string, Function*>::iterator       FI;

Executable::Executable(Process proc, string prog_name) : 
  proc(proc), 
  name(prog_name) 
{
  AisStatus sts;
  
  SourceObj prog  = proc.get_program_object();

  // look for modules
  for (int c = 0 ; c < prog.child_count(); c++) 
    {
      SourceObj modobj = prog.child(c);

      vector<char> mod_name(modobj.module_name_length());
      modobj.module_name(&mod_name[0], mod_name.size());

      Module* mod = new Module();
      modm[&mod_name[0]] = mod;
      
      mod->name  = &mod_name[0];
      mod->begln = modobj.line_start();

      // cerr << "MODULE " << mod->name << ":" << mod->begln << endl;

      sts = modobj.bexpand(proc);

      if (sts.status() != ASC_success) 
	{
	  proc.bdestroy();
	  elg_error_msg("bexpand failed: %s", sts.status_name());
	}

      // look for functions
      for ( int d = 0; d < modobj.child_count(); d++) 
	{

	  SourceObj funcobj = modobj.child(d);

	  vector<char> func_name(funcobj.get_demangled_name_length());
	  funcobj.get_demangled_name(&func_name[0], func_name.size());

	  if ( strcmp(&func_name[0], "") == 0 )
	    continue;

	  Function* func = new Function();
	  funcm[&func_name[0]] = func;

	  func->name      = &func_name[0];
	  func->begln     = funcobj.line_start();
	  func->endln     = funcobj.line_end();
	  func->mod       = mod;
	  func->ismpicoll = func->is_mpi_coll(); // for performance

	  // cerr << "FUNCTION " << func->name << ":" << func->begln << ":" << func->endln << endl;
	  
	  Callsite* csite = NULL;
	  
	  // look for call sites
	  for ( int e = 0; e < funcobj.exclusive_point_count(); e++) {
	  
	    InstPoint point = funcobj.exclusive_point(e);

	      if ( point.get_type() == IPT_function_entry )
		{
		  func->entry = point;
		}
	      else if ( point.get_type() == IPT_function_exit )
		{
		  func->exit = point;		  
		}
	      else if ( point.get_type() == IPT_function_call )
		{    
		  vector<char> callee_name(point.get_demangled_name_length());
		  point.get_demangled_name(&callee_name[0], callee_name.size());
		  		  
		    if ( point.get_location() == IPL_before )
		    {
		      csite = new Callsite();
		      csitev.push_back(csite);

		      csite->mod         = mod;
		      csite->line        = point.get_line();
		      csite->caller      = func;
		      csite->callee_name = &callee_name[0];
		      csite->before      = point;
		    }
		    else if ( point.get_location() == IPL_after )
		    {
		      csite->after = point;
		    }
		}
	    }	  
	}
    }

  // resolve callee relationships, remove invalid call sites 

  vector<Callsite*> tmp; 
  
  tmp.clear();
  for ( int i = 0; i < csitev.size(); i++ )
    if ( funcm.find(csitev[i]->callee_name) != funcm.end() )
      {
	csitev[i]->callee = funcm[csitev[i]->callee_name];
	tmp.push_back(csitev[i]);
      }
    else
      {
	delete csitev[i];
      }
  csitev = tmp;

  // delete non-user entities

  // classify based on line-number information 

  for ( MCI it = modm.begin(); it != modm.end(); it++ )
    {
      Module* mod = it->second;
      mod->isuser = false;
      mod->isleaf = false;
      if ( mod->begln > 0 )
	mod->isuser = true;
    }
  for ( FCI it = funcm.begin(); it != funcm.end(); it++ )
    {
      Function* func = it->second; 
      func->isuser = false;
      func->isleaf = false;
      if ( func->begln > 0 ) 
	func->isuser = true;
    }
  for ( int i = 0; i < csitev.size(); i++ )
    if ( csitev[i]->line > 0 )
      csitev[i]->isuser = true;
    
  // identify non-user functions including their modules that are
  // directly called from user functions and that are either MPI or
  // OpenMP library calls (leaf entities)

  for ( int i = 0; i < csitev.size(); i++ )
    if ( csitev[i]->isuser && !csitev[i]->callee->isuser )
      {
	if ( csitev[i]->callee->is_mpi() || csitev[i]->callee->is_omp() )
	  {
	    csitev[i]->callee->isleaf = true;      // callee function
	    csitev[i]->callee->mod->isleaf = true; // callee module
	  }
	else
	  {
	    // mark as non-user call site because it points neither to
	    // a user nor to an MPI/OpenMP function
	    csitev[i]->isuser = false;
	  }
      }

  // delete non-user and non-leaf entities

  // modules
  for ( MI it = modm.begin(); it != modm.end(); it++ )
    if ( it->second->isuser || it->second->isleaf )
      {
	modv.push_back(it->second);
      }
    else
      {
	delete it->second;
      }
  modm.clear();
  for ( int i = 0; i < modv.size(); i++)
    modm[modv[i]->name] = modv[i];

  // functions
  for ( FI it = funcm.begin(); it != funcm.end(); it++ )
    if ( it->second->isuser || it->second->isleaf )
      {
	funcv.push_back(it->second);
	if ( it->second->isleaf )
	  it->second->csitev.clear();
      }
    else
      {
	delete it->second;
      }
  funcm.clear();
  for ( int i = 0; i < funcv.size(); i++)
    funcm[funcv[i]->name] = funcv[i];

  // call sites
  tmp.clear();
  for ( int i = 0; i < csitev.size(); i++ )
    if ( csitev[i]->isuser )
      tmp.push_back(csitev[i]);
    else
      delete csitev[i];
  csitev = tmp;

  // add functions to their module's function vector

  for ( int i = 0; i < funcv.size(); i++ )
    funcv[i]->mod->funcv.push_back(funcv[i]);

  // add call sites to their function's call-site  vector

  for ( int i = 0; i < csitev.size(); i++ )
    csitev[i]->caller->csitev.push_back(csitev[i]);

  // assign ids

  for (int i = 0; i < modv.size(); i++)
    modv[i]->id = i;
  for (int i = 0; i < funcv.size(); i++)
    funcv[i]->id = i;
  for (int i = 0; i < csitev.size(); i++)
    csitev[i]->id = i;
}

Executable::~Executable()
{
  for (int i = 0; i < modv.size(); i++)
    delete modv[i];
  for (int i = 0; i < funcv.size(); i++)
    delete funcv[i];
  for (int i = 0; i < csitev.size(); i++)
    delete csitev[i];
}

Module*   
Executable::get_mod(string name) const
{
  MCI result = modm.find(name);

  if ( result != modm.end() )
    return result->second;
  else
    return NULL;
}

Function* 
Executable::get_func(string name) const 
{
  FCI result = funcm.find(name);

  if ( result != funcm.end() )
    return result->second;
  else
    return NULL;
}
 

ostream& operator<<(ostream& out, const Executable& exec)
{
  out << "PROGRAM" << " " 
      << exec.get_name() << " "
      << exec.get_num_mods() << " "
      << exec.get_num_funcs() << " "
      << exec.get_num_csites() << endl;

  for (int i = 0; i < exec.get_num_mods(); i++)
    out << *exec.get_mod(i);
  for (int i = 0; i < exec.get_num_funcs(); i++)
    out << *exec.get_func(i);
  for (int i = 0; i < exec.get_num_csites(); i++)
    out << *exec.get_csite(i);

  return out;
}
