/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "elg_trc.h"
#include "elg_mpicom.h"
#include "elg_error.h"

#include <stdlib.h>

/* 
 *-----------------------------------------------------------------------------
 *
 * Communicator management
 *
 *-----------------------------------------------------------------------------
 */

#define ELG_MAX_COMM   50

struct ElgWorld
{
  MPI_Group group;
  int size;
  int size_grpv;
  int* ranks;
};

struct ElgWorld world;

int elg_rank_to_pe(int rank, MPI_Comm comm)
{
  MPI_Group group;
  int global_rank;
  int inter;

  MPI_Comm_test_inter(comm, &inter);
  if ( inter )
    MPI_Comm_remote_group(comm, &group);
  else
    MPI_Comm_group(comm, &group);

  MPI_Group_translate_ranks(group, 1, &rank, world.group, &global_rank);
  MPI_Group_free(&group);
  return global_rank;
}

struct ElgComm
{
  MPI_Comm        comm;
  elg_ui4          cid;
};

static int currcid = 1;
static int last_comm = 0;
static int* ranks;
static struct ElgComm comms[ELG_MAX_COMM];
static unsigned char* grpv;


void elg_comm_init()
{
  int i;

  PMPI_Comm_group(MPI_COMM_WORLD, &world.group);
  PMPI_Group_size(world.group, &world.size);
  world.size_grpv = world.size / 8 + (world.size % 8 ? 1 : 0);

  world.ranks  = calloc(world.size, sizeof(int));
  for (i = 0; i < world.size; i++)
    world.ranks[i] = i;  

  ranks  = calloc(world.size, sizeof(int));
  grpv = calloc(world.size_grpv, sizeof(unsigned char));
}

void elg_comm_finalize()
{
  PMPI_Group_free(&world.group);

  free(world.ranks);
  free(ranks);
  free(grpv);
}

void elg_comm_create(MPI_Comm comm)
{
  int i;
  MPI_Group group;
  
  /* ask for group of comm */
  PMPI_Comm_group(comm, &group);

  /* translate ranks */
  PMPI_Group_translate_ranks(world.group, world.size, world.ranks, group, ranks);
  
  /* initialize grpv */
  for (i = 0; i < world.size_grpv; i++)
    grpv[i] = 0;

  /* which process in MPI_COMM_WORLD is member of comm */
  for (i = 0; i < world.size; i++)
    if (ranks[i] != MPI_UNDEFINED)
      grpv[i / 8] |= (1 << (i % 8));

  /* write mpi_com record */
  elg_mpi_com(currcid, world.size_grpv, grpv);

  /* enter comm in comms[] arrray */

  comms[last_comm].comm = comm;
  comms[last_comm].cid  = currcid++; 
  
  if (++last_comm > ELG_MAX_COMM) 
    elg_error_msg("To many communicators");

  /* clean up */
  PMPI_Group_free(&group); 
}

void elg_comm_free(MPI_Comm comm)
{
  int myrank;

  PMPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  
  if (last_comm == 1 && comms[0].comm == comm)
    {
      last_comm = 0;
    }
  else if (last_comm > 1)
    {
      int i = 0;
      
      while(i < last_comm && comms[i].comm != comm)
	i++;
      
      if (i < last_comm--)
	{
	  PMPI_Comm_rank(MPI_COMM_WORLD, &myrank);
	  comms[i] = comms[last_comm];
	}
      else
	elg_error_msg("Cannot find communicator");
    }
  else
    {
      elg_error_msg("Cannot find communicator");
    }
}

elg_ui4 elg_comm_id(MPI_Comm comm)
{
  int i = 0;

  while(i < last_comm && comms[i].comm != comm)
    i++;
  
  if (i < last_comm)
    return comms[i].cid;
  else 
    {
      elg_error_msg("Cannot find communicator");  
      return -1;
    }
}
























