/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "elg_trc.h"
#include "elg_mpicom.h"
#include "elg_mpireg.h"
#include "elg_mpireq.h"
#include "elg_error.h"

#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

/* 
 *-----------------------------------------------------------------------------
 *
 * Init and finalize
 *
 *-----------------------------------------------------------------------------
 */

int elg_mpi_notrace = 0;

#define IS_TRACE_ON    !elg_mpi_notrace 
#define TRACE_OFF()     elg_mpi_notrace = 1
#define TRACE_ON()      elg_mpi_notrace = 0
#define MAX_REQUESTS    1024

/* -- MPI_Init -- */

int MPI_Init( int *argc, char ***argv )
{
  int returnVal, numprocs, i;
  unsigned char* grpv;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
      elg_open();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_INIT]); 
#endif          
      returnVal = PMPI_Init(argc, argv);
      
 
      /* initialize trace file generation */
      elg_mpi_init();

      PMPI_Comm_size(MPI_COMM_WORLD, &numprocs);      
      
      /* write mpi_com record for MPI_COMM_WORLD */
      grpv = calloc(numprocs / 8 + (numprocs % 8 ? 1 : 0), sizeof(unsigned char));
      for (i = 0; i < numprocs; i++)
	grpv[i / 8] |= (1 << (i % 8));
      elg_mpi_com(0, numprocs / 8 + (numprocs % 8 ? 1 : 0), grpv);
      free(grpv);
            
      /* initialize communicator management */
      elg_comm_init();

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      returnVal = PMPI_Init(argc, argv);
    }
  
  return returnVal;
}

/* -- MPI_Finalize -- */

int MPI_Finalize()
{
  int returnVal;
  
  if (IS_TRACE_ON)
    {
      TRACE_OFF(); 
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_FINALIZE]); 
#endif    
      /* finalize communicator and request management */
      elg_comm_finalize();
      elg_request_finalize();
      
      /* finalize trace file generation */
      elg_mpi_finalize();
      
      returnVal = PMPI_Finalize();

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      returnVal = PMPI_Finalize();
    }
  
  return returnVal;
}

/* 
 *-----------------------------------------------------------------------------
 *
 * Communicator management
 *
 *-----------------------------------------------------------------------------
 */

/* ------- Constructors ------- */  

int MPI_Comm_dup( MPI_Comm comm, 
                  MPI_Comm* newcomm )
{
  int result;
  
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_COMM_DUP]); 
#endif
      result = PMPI_Comm_dup(comm, newcomm);
      elg_comm_create(*newcomm);  

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Comm_dup(comm, newcomm);
    }
  
  return result;
}

int MPI_Comm_create( MPI_Comm comm, 
                     MPI_Group group, 
                     MPI_Comm* newcomm )
{
  int result;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_COMM_CREATE]); 
#endif
      result = PMPI_Comm_create(comm, group, newcomm);
      if ( *newcomm != MPI_COMM_NULL)
        elg_comm_create(*newcomm);  

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Comm_create(comm, group, newcomm);
    }

  return result;
}

int MPI_Comm_split( MPI_Comm comm, 
                    int color, 
                    int key, 
                    MPI_Comm* newcomm )
{
  int result;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_COMM_SPLIT]); 
#endif
      result = PMPI_Comm_split(comm, color, key, newcomm);
      if ( *newcomm != MPI_COMM_NULL)
        elg_comm_create(*newcomm);  

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Comm_split(comm, color, key, newcomm);
    }

  return result;
}


/* 
 *-----------------------------------------------------------------------------
 *
 * Cartesian Toplogy functions
 *
 *-----------------------------------------------------------------------------
 */

int MPI_Cart_create( MPI_Comm comm_old,
                     int ndims,
                     int* dims,
                     int* periods,
                     int reorder,
                     MPI_Comm* comm_cart) 
{
  int result, i;
  int topid, my_rank; 
  int *coordv;

  elg_ui1  undims, *uperiodv;
  elg_ui4  *udimv, *ucoordv;
    
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST
      elg_enter(elg_mpi_regid[ELG__MPI_CART_CREATE]);
#endif
      result = PMPI_Cart_create(comm_old, ndims, dims, periods, reorder, comm_cart);
      if ( *comm_cart != MPI_COMM_NULL)
	{
	  elg_comm_create(*comm_cart);
	  
	  /* get the topid */
	  topid = elg_comm_id(*comm_cart);
	  
	  /* find the rank of the calling process */
	  PMPI_Comm_rank(*comm_cart, &my_rank);	
	  
	  /* assign the parameters to write ELG_CART_TOPOLOGY */
	  undims   =  (elg_ui1) ndims;
	  udimv = calloc(ndims, sizeof(elg_ui4));
	  if ( udimv == NULL ) 
	    elg_error();
	  uperiodv = calloc(ndims, sizeof(elg_ui1));
	  if ( uperiodv == NULL ) 
	    elg_error();
	  for ( i = 0 ; i < ndims ; i++ )
	    {
	      udimv[i]    =  (elg_ui4) dims[i];
	      uperiodv[i] =  (elg_ui1) periods[i];
	    }
	 	     
	  /*  write the definition record ELG_CART_TOPOLOGY */
	  elg_def_cart(topid, undims, udimv, uperiodv);

	  /* allocate space for coordv and ucoordv */
	  coordv = calloc(ndims, sizeof(int));
	  if ( coordv == NULL ) 
	    elg_error();
	  ucoordv = calloc(ndims, sizeof(elg_ui4));
	  if ( ucoordv == NULL ) 
	    elg_error();
				     
	  
	  /* get the coordinates of the root process in coordv */
	  PMPI_Cart_coords(*comm_cart, my_rank, ndims, coordv); 

	  /*  asssign the parameters to write ELG_CART_COORDS */
	  for ( i = 0 ; i < ndims ; i++ )
	    {
	      ucoordv[i] = (elg_ui4) coordv[i];
	    }
	
	  /* write the definition record ELG_CART_COORDS */
	  elg_def_coords(topid, undims, ucoordv);	   	 

	  free(udimv);
	  free(uperiodv);
	  free(ucoordv);
	}

#ifndef ELG_CSITE_INST
      elg_exit();
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Cart_create(comm_old, ndims, dims, periods, reorder, comm_cart);
    }
   
  return result;
}


int MPI_Cart_sub ( MPI_Comm comm, 
                   int *rem_dims, 
                   MPI_Comm *newcomm)
{
  int result;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_CART_SUB]); 
#endif
      result = PMPI_Cart_sub(comm, rem_dims, newcomm);
      if ( *newcomm != MPI_COMM_NULL)
        elg_comm_create(*newcomm);  

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Cart_sub(comm, rem_dims, newcomm );
    }

  return result;
}



int MPI_Graph_create( MPI_Comm comm_old,
                      int nnodes,
                      int* index,
                      int* edges,
                      int reorder,
                      MPI_Comm* comm_graph) 
{
  int result;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_GRAPH_CREATE]); 
#endif
      result = PMPI_Graph_create(comm_old, nnodes, index, edges, reorder, comm_graph);
      if ( *comm_graph != MPI_COMM_NULL)
        elg_comm_create(*comm_graph);  

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Graph_create(comm_old, nnodes, index, edges, reorder, comm_graph);
    }

  return result;
}

int MPI_Intercomm_create (MPI_Comm local_comm, 
                          int local_leader, 
                          MPI_Comm peer_comm, 
                          int remote_leader, 
                          int tag, 
                          MPI_Comm *newintercomm)

{
  int result;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_INTERCOMM_CREATE]); 
#endif
      result = PMPI_Intercomm_create(local_comm, local_leader, peer_comm, 
                                     remote_leader, tag, newintercomm);
      if ( *newintercomm != MPI_COMM_NULL)
        elg_comm_create(*newintercomm);  

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Intercomm_create(local_comm, local_leader, peer_comm, 
                                     remote_leader, tag, newintercomm);
    }

  return result;
}

int MPI_Intercomm_merge (MPI_Comm intercomm, 
                         int high, 
                         MPI_Comm *newcomm)
{
  int result;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_INTERCOMM_MERGE]); 
#endif
      result = PMPI_Intercomm_merge(intercomm, high, newcomm);
      if ( *newcomm != MPI_COMM_NULL)
        elg_comm_create(*newcomm);  

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Intercomm_merge ( intercomm, high, newcomm);
    }

  return result;
}


/* ------- Destructors ------- */  

int MPI_Comm_free( MPI_Comm* comm )
{
  int result;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_COMM_FREE]); 
#endif
      elg_comm_free(*comm);
      result = PMPI_Comm_free(comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Comm_free(comm);
    }

  return result;
}


/* 
 *-----------------------------------------------------------------------------
 *
 * Point-to-point communication
 *
 *-----------------------------------------------------------------------------
 */

/* ------- Synchronous ------- */


/* -- MPI_Send -- */

int MPI_Send( void* buf, 
	      int count, 
	      MPI_Datatype datatype, 
	      int dest, 
	      int tag, 
	      MPI_Comm comm )
{
  int result, sz;
  
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_SEND]); 
#endif
      if (dest != MPI_PROC_NULL)
	{
	  PMPI_Type_size(datatype, &sz);
	  elg_mpi_send(ELG_RANK_TO_PE(dest, comm), ELG_COMM_ID(comm), 
		       tag, count * sz);
	}
      result = PMPI_Send(buf, count, datatype, dest, tag, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif      
      TRACE_ON();
    }
  else
    {
      result = PMPI_Send(buf, count, datatype, dest, tag, comm);
    }

  return result;
}

/* -- MPI_Bsend -- */

int MPI_Bsend( void* buf, 
	      int count, 
	      MPI_Datatype datatype, 
	      int dest, 
	      int tag, 
	      MPI_Comm comm )
{
  int result, sz;
  
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_BSEND]); 
#endif
      if (dest != MPI_PROC_NULL)
	{
	  PMPI_Type_size(datatype, &sz);
	  elg_mpi_send(ELG_RANK_TO_PE(dest, comm), ELG_COMM_ID(comm), 
		       tag, count * sz);
	}
      result = PMPI_Bsend(buf, count, datatype, dest, tag, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif      
      TRACE_ON();
    }
  else
    {
      result = PMPI_Bsend(buf, count, datatype, dest, tag, comm);
    }

  return result;
}

/* -- MPI_Rsend -- */

int MPI_Rsend( void* buf,
               int count,
               MPI_Datatype datatype,
               int dest,
               int tag,
               MPI_Comm comm)
{
  int result, sz;
  
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_RSEND]); 
#endif
      if (dest != MPI_PROC_NULL)
	{
	  PMPI_Type_size(datatype, &sz);
	  elg_mpi_send(ELG_RANK_TO_PE(dest, comm), ELG_COMM_ID(comm), 
		       tag, count * sz);
	}
      result = PMPI_Rsend(buf, count, datatype, dest, tag, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif      
      TRACE_ON();
    }
  else
    {
      result = PMPI_Rsend(buf, count, datatype, dest, tag, comm);
    }

  return result;
}

/* -- MPI_Ssend -- */

int MPI_Ssend( void* buf, 
	      int count, 
	      MPI_Datatype datatype, 
	      int dest, 
	      int tag, 
	      MPI_Comm comm )
{
  int result, sz;
  
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_SSEND]); 
#endif
      if (dest != MPI_PROC_NULL)
	{
	  PMPI_Type_size(datatype, &sz);
	  elg_mpi_send(ELG_RANK_TO_PE(dest, comm), ELG_COMM_ID(comm), 
		       tag, count * sz);
	}
      result = PMPI_Ssend(buf, count, datatype, dest, tag, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif      
      TRACE_ON();
    }
  else
    {
      result = PMPI_Ssend(buf, count, datatype, dest, tag, comm);
    }

  return result;
}

/* -- MPI_Recv -- */

int MPI_Recv( void* buf, 
	      int count, 
	      MPI_Datatype datatype, 
	      int source, int tag, 
	      MPI_Comm comm, 
	      MPI_Status* status )
{
  int result;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_RECV]); 
#endif
      result = PMPI_Recv(buf, count, datatype, source, tag, comm, status);
      
      if (source != MPI_PROC_NULL && result == MPI_SUCCESS)
	elg_mpi_recv(ELG_RANK_TO_PE(status->MPI_SOURCE, comm),
                     ELG_COMM_ID(comm), status->MPI_TAG);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Recv(buf, count, datatype, source, tag, comm, status);
    }

  return result;
}


/* -- MPI_Sendrecv -- */

int MPI_Sendrecv(void* sendbuf, 
		 int sendcount, 
		 MPI_Datatype sendtype, 
		 int dest,   
		 int sendtag,
		 void* recvbuf, 
		 int recvcount, 
		 MPI_Datatype recvtype, 
		 int source, 
		 int recvtag,
		 MPI_Comm comm, 
		 MPI_Status* status )
{
  int result, sendsz;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_SENDRECV]); 
#endif
      if (dest != MPI_PROC_NULL)
        {
	  PMPI_Type_size(sendtype, &sendsz);
	  elg_mpi_send(ELG_RANK_TO_PE(dest, comm), ELG_COMM_ID(comm),
		       sendtag, sendcount * sendsz);
        }
      result = PMPI_Sendrecv(sendbuf, sendcount, sendtype, dest,   sendtag,
			     recvbuf, recvcount, recvtype, source, recvtag,
			     comm, status);
      if (source != MPI_PROC_NULL && result == MPI_SUCCESS)
        {
	  elg_mpi_recv(ELG_RANK_TO_PE(status->MPI_SOURCE, comm),
	               ELG_COMM_ID(comm), status->MPI_TAG);
	}

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Sendrecv(sendbuf, sendcount, sendtype, dest,   sendtag,
			     recvbuf, recvcount, recvtype, source, recvtag,
			     comm, status);
    }

  return result;
}

/* -- MPI_Sendrecv_replace -- */

int MPI_Sendrecv_replace(void* buf, 
		         int count, 
		         MPI_Datatype datatype, 
		         int dest,   
		         int sendtag,
		         int source, 
		         int recvtag,
		         MPI_Comm comm, 
		         MPI_Status* status )
{
  int result, sz;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_SENDRECV_REPLACE]); 
#endif
      if (dest != MPI_PROC_NULL)
        {
	  PMPI_Type_size(datatype, &sz);
	  elg_mpi_send(ELG_RANK_TO_PE(dest, comm),
		       ELG_COMM_ID(comm),
		       sendtag,
		       count * sz);
        }
      result = PMPI_Sendrecv_replace(buf, count, datatype, dest,
	                             sendtag, source, recvtag,
				     comm, status);
      if (source != MPI_PROC_NULL && result == MPI_SUCCESS)
        {
	  elg_mpi_recv(ELG_RANK_TO_PE(status->MPI_SOURCE, comm),
	               ELG_COMM_ID(comm), status->MPI_TAG);
	}

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Sendrecv_replace(buf, count, datatype, dest,
	                             sendtag, source, recvtag,
				     comm, status);
    }

  return result;
}

/* ------- Aynchronous ------- */

/* -- MPI_Isend -- */

int MPI_Isend( void* buf, 
	       int count, 
	       MPI_Datatype datatype, 
	       int dest, 
	       int tag, 
	       MPI_Comm comm, 
	       MPI_Request* request )
{
  int result, sz;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_ISEND]); 
#endif
      if (dest != MPI_PROC_NULL)
	{
	  PMPI_Type_size(datatype, &sz);
	  elg_mpi_send(ELG_RANK_TO_PE(dest, comm), ELG_COMM_ID(comm), 
		       tag, count * sz);
	}
      result = PMPI_Isend(buf, count, datatype, dest, tag, comm, request);
      if (dest != MPI_PROC_NULL && result == MPI_SUCCESS)
        elg_request_create(*request, SEND_REQ, tag, dest, count*sz, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Isend(buf, count, datatype, dest, tag, comm, request);
    }

  return result;
}


/* -- MPI_Irecv -- */

int MPI_Irecv( void* buf, 
	       int count, 
	       MPI_Datatype datatype, 
	       int source, 
	       int tag, 
	       MPI_Comm comm, 
	       MPI_Request* request )
{
  int result, sz;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_IRECV]); 
#endif
      PMPI_Type_size(datatype, &sz);
      
      result = PMPI_Irecv(buf, count, datatype, source, tag, comm, request);
      if (source != MPI_PROC_NULL && result == MPI_SUCCESS)
        elg_request_create(*request, RECV_REQ, tag, 0, count * sz, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Irecv(buf, count, datatype, source, tag, comm, request);
    }

  return result;
}

/* -- MPI_Ibsend -- */

int MPI_Ibsend( void* buf,
                int count,
                MPI_Datatype datatype,
                int dest,
                int tag,
                MPI_Comm comm,
                MPI_Request *request)
{
  int result, sz;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_IBSEND]); 
#endif
      if (dest != MPI_PROC_NULL)
	{
	  PMPI_Type_size(datatype, &sz);
	  elg_mpi_send(ELG_RANK_TO_PE(dest, comm), ELG_COMM_ID(comm), 
		       tag, count * sz);
	}
      result = PMPI_Ibsend(buf, count, datatype, dest, tag, comm, request);
      if (dest != MPI_PROC_NULL && result == MPI_SUCCESS)
        elg_request_create(*request, SEND_REQ, tag, dest, count*sz, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Ibsend(buf, count, datatype, dest, tag, comm, request);
    }

  return result;
}

/* -- MPI_Issend -- */

int MPI_Issend( void* buf,
                int count,
                MPI_Datatype datatype,
                int dest,
                int tag,
                MPI_Comm comm,
                MPI_Request *request)
{
  int result, sz;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_ISSEND]); 
#endif
      if (dest != MPI_PROC_NULL)
	{
	  PMPI_Type_size(datatype, &sz);
	  elg_mpi_send(ELG_RANK_TO_PE(dest, comm), ELG_COMM_ID(comm), 
		       tag, count * sz);
	}
      result = PMPI_Issend(buf, count, datatype, dest, tag, comm, request);
      if (dest != MPI_PROC_NULL && result == MPI_SUCCESS)
        elg_request_create(*request, SEND_REQ, tag, dest, count*sz, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Issend(buf, count, datatype, dest, tag, comm, request);
    }

  return result;
}

/* -- MPI_Irsend -- */

int MPI_Irsend( void* buf,
                int count,
                MPI_Datatype datatype,
                int dest,
                int tag,
                MPI_Comm comm,
                MPI_Request *request)
{
  int result, sz;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_IRSEND]); 
#endif
      if (dest != MPI_PROC_NULL)
	{
	  PMPI_Type_size(datatype, &sz);
	  elg_mpi_send(ELG_RANK_TO_PE(dest, comm), ELG_COMM_ID(comm), 
		       tag, count * sz);
	}
      result = PMPI_Irsend(buf, count, datatype, dest, tag, comm, request);
      if (dest != MPI_PROC_NULL)
        elg_request_create(*request, SEND_REQ, tag, dest, count*sz, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Irsend(buf, count, datatype, dest, tag, comm, request);
    }

  return result;
}

/* -- MPI_Wait -- */

int MPI_Wait( MPI_Request* request, 
	      MPI_Status* status)
{
  int result;

  if (IS_TRACE_ON)
    {
      struct ElgRequest* orig_req;

      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_WAIT]); 
#endif
      orig_req = elg_request_get(*request);
      result = PMPI_Wait(request, status);
      elg_check_request(orig_req, status);
#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif      
      TRACE_ON();
    }
  else
    {
      result = PMPI_Wait(request, status);
    }


  return result;
}

/* -- MPI_Waitall -- */

int MPI_Waitall( int count, 
		 MPI_Request* requests, 
		 MPI_Status* statuses )
{
  int result, i;
  struct ElgRequest* orig_req;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_WAITALL]); 
#endif
      elg_save_request_array(requests, count);
      result = PMPI_Waitall(count, requests, statuses);
      for (i = 0; i < count; i++)
        {
	  orig_req = elg_saved_request_get(i);
          elg_check_request(orig_req, &(statuses[i]));
        }
#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Waitall(count, requests, statuses);
    }

  return result;
}

/* -- MPI_Waitany -- */

int MPI_Waitany( int count, 
		 MPI_Request* requests, 
		 int* index, 
		 MPI_Status* status )
{
  int result;
  struct ElgRequest* orig_req;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_WAITANY]); 
#endif
      elg_save_request_array(requests, count);
      result = PMPI_Waitany(count, requests, index, status);
      orig_req = elg_saved_request_get(*index);
      elg_check_request(orig_req, status);
#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Waitany(count, requests, index, status);
    }


  return result;
}

/* -- MPI_Waitsome -- */

int MPI_Waitsome(int incount,
		 MPI_Request *array_of_requests,
		 int *outcount,
		 int *array_of_indices,
		 MPI_Status *array_of_statuses )
{
  int result, i;
  struct ElgRequest* orig_req;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_WAITSOME]); 
#endif
      elg_save_request_array(array_of_requests, incount);
      result = PMPI_Waitsome(incount, array_of_requests, outcount,
	                     array_of_indices, array_of_statuses );
      for (i=0; i<*outcount; ++i)
        {
          orig_req = elg_saved_request_get(array_of_indices[i]);
          elg_check_request(orig_req, &(array_of_statuses[i]));
        }
#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Waitsome(incount, array_of_requests, outcount,
	                     array_of_indices, array_of_statuses );
    }

  return result;
}

/* -- MPI_Test -- */

int MPI_Test( MPI_Request* request, 
	      int* flag, 
	      MPI_Status* status )
{
  int result;
  struct ElgRequest* orig_req;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_TEST]); 
#endif
      orig_req = elg_request_get(*request);
      result = PMPI_Test(request, flag, status);
      if (*flag) elg_check_request(orig_req, status);
#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Test(request, flag, status);
    }

  return result;
}

/* -- MPI_Testany -- */

int MPI_Testany( int count,
                 MPI_Request *array_of_requests,
		 int *index,
		 int *flag,
		 MPI_Status *status )
{
  int result;
  struct ElgRequest* orig_req;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_TESTANY]); 
#endif
      elg_save_request_array(array_of_requests, count);
      result = PMPI_Testany( count, array_of_requests, index, flag, status );
      if (*flag && *index != MPI_UNDEFINED)
        {
          orig_req = elg_saved_request_get(*index);
          elg_check_request(orig_req, status);
        }
#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Testany( count, array_of_requests, index, flag, status );
    }

  return result;
}

/* -- MPI_Testall -- */

int MPI_Testall( int count,
                 MPI_Request *array_of_requests,
		 int *flag,
		MPI_Status *array_of_statuses )
{
  int result, i;
  struct ElgRequest* orig_req;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_TESTALL]); 
#endif
      elg_save_request_array(array_of_requests, count);
      result = PMPI_Testall(count, array_of_requests, flag, array_of_statuses );
      if (*flag)
        {
	  for (i = 0; i < count; i++)
            {
	      orig_req = elg_saved_request_get(i);
              elg_check_request(orig_req, &(array_of_statuses[i]));
            }
        }
#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Testall(count, array_of_requests, flag, array_of_statuses );
    }

  return result;
}

/* -- MPI_Testsome -- */

int MPI_Testsome(int incount,
		 MPI_Request *array_of_requests,
		 int *outcount,
		 int *array_of_indices,
		 MPI_Status *array_of_statuses )
{
  int result, i;
  struct ElgRequest* orig_req;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_TESTSOME]); 
#endif
      elg_save_request_array(array_of_requests, incount);
      result = PMPI_Testsome( incount, array_of_requests, outcount,
	                      array_of_indices, array_of_statuses );
      for (i=0; i<*outcount; ++i)
        {
          orig_req = elg_saved_request_get(array_of_indices[i]);
          elg_check_request(orig_req, &(array_of_statuses[i]));
        }
#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Testsome( incount, array_of_requests, outcount,
	                      array_of_indices, array_of_statuses );
    }

  return result;
}

/* ------- Persistent requests ------- */

/* -- MPI_Send_init -- */

int MPI_Send_init( void* buf, 
		   int count, 
		   MPI_Datatype datatype, 
		   int dest, 
		   int tag, 
		   MPI_Comm comm,
		   MPI_Request* request )
{
  int result, sz;
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_SEND_INIT]); 
#endif
      PMPI_Type_size(datatype, &sz);
      
      result = PMPI_Send_init(buf, count, datatype, dest, tag, comm, request);
      if (dest != MPI_PROC_NULL && result == MPI_SUCCESS)
        elg_request_create(*request, SEND_PERS_REQ, tag, dest, count*sz, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Send_init(buf, count, datatype, dest, tag, comm, request);
    }

  return result;
}

/* -- MPI_Recv_init -- */

int MPI_Recv_init( void* buf, 
		   int count, 
		   MPI_Datatype datatype, 
		   int source, 
		   int tag, 
		   MPI_Comm comm,
		   MPI_Request* request )
{
  int result, sz;
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_RECV_INIT]); 
#endif
      PMPI_Type_size(datatype, &sz);
      
      result = PMPI_Recv_init(buf, count, datatype, source, tag, comm, request);
      if (source != MPI_PROC_NULL && result == MPI_SUCCESS)
        elg_request_create(*request, RECV_PERS_REQ, tag, source,
                           count * sz, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Recv_init(buf, count, datatype, source, tag, comm, request);
    }


  return result;
}

/* -- MPI_Bsend_init -- */

int MPI_Bsend_init( void* buf, 
		   int count, 
		   MPI_Datatype datatype, 
		   int dest, 
		   int tag, 
		   MPI_Comm comm,
		   MPI_Request* request )
{
  int result, sz;
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_BSEND_INIT]); 
#endif
      PMPI_Type_size(datatype, &sz);
      
      result = PMPI_Bsend_init(buf, count, datatype, dest, tag, comm, request);
      if (dest != MPI_PROC_NULL && result == MPI_SUCCESS)
        elg_request_create(*request, SEND_PERS_REQ, tag, dest, count*sz, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Bsend_init(buf, count, datatype, dest, tag, comm, request);
    }

  return result;
}

/* -- MPI_Ssend_init -- */

int MPI_Ssend_init( void* buf, 
		   int count, 
		   MPI_Datatype datatype, 
		   int dest, 
		   int tag, 
		   MPI_Comm comm,
		   MPI_Request* request )
{
  int result, sz;
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_SSEND_INIT]); 
#endif
      PMPI_Type_size(datatype, &sz);
      
      result = PMPI_Ssend_init(buf, count, datatype, dest, tag, comm, request);
      if (dest != MPI_PROC_NULL && result == MPI_SUCCESS)
        elg_request_create(*request, SEND_PERS_REQ, tag, dest, count*sz, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Ssend_init(buf, count, datatype, dest, tag, comm, request);
    }

  return result;
}

/* -- MPI_Rsend_init -- */

int MPI_Rsend_init( void* buf, 
		   int count, 
		   MPI_Datatype datatype, 
		   int dest, 
		   int tag, 
		   MPI_Comm comm,
		   MPI_Request* request )
{
  int result, sz;
  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_RSEND_INIT]); 
#endif
      PMPI_Type_size(datatype, &sz);
      
      result = PMPI_Rsend_init(buf, count, datatype, dest, tag, comm, request);
      if (dest != MPI_PROC_NULL && result == MPI_SUCCESS)
        elg_request_create(*request, SEND_PERS_REQ, tag, dest, count*sz, comm);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Rsend_init(buf, count, datatype, dest, tag, comm, request);
    }

  return result;
}

/* -- MPI_Start -- */

int MPI_Start( MPI_Request* request )
{
  int result;
  
  if (IS_TRACE_ON)
    {
      struct ElgRequest* req;

      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_START]); 
#endif
      req = elg_request_get(*request);
      if (req && req->type == SEND_PERS_REQ && req->dest != MPI_PROC_NULL)
	elg_mpi_send(ELG_RANK_TO_PE(req->dest, req->comm), 
		     ELG_COMM_ID(req->comm), req->tag,  req->bytes);

      result = PMPI_Start(request);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Start(request);
    }


  return result;
}

/* -- MPI_Startall -- */

int MPI_Startall( int count,
                  MPI_Request *array_of_requests )
{
  int result, i;

  if (IS_TRACE_ON)
    {
      struct ElgRequest* req;
      MPI_Request *request;

      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_STARTALL]); 
#endif
      for (i=0; i < count-1; ++i)
        {
           request=&array_of_requests[i];
           req = elg_request_get(*request);
           /* if receive request, write receive trace record */
           if (req &&
               (req->type == SEND_PERS_REQ || req->type == SEND_REQ) &&
	       (req->dest != MPI_PROC_NULL) )
	    elg_mpi_send(ELG_RANK_TO_PE(req->dest, req->comm),
		       ELG_COMM_ID(req->comm), req->tag, req->bytes);
      }
      result = PMPI_Startall( count, array_of_requests );

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
  }
  else
  {
	result = PMPI_Startall( count, array_of_requests );
  }
  return result;
}

/* -- MPI_Request_free -- */

int MPI_Request_free( MPI_Request* request )
{
  int result;

  if (IS_TRACE_ON)
    {
      struct ElgRequest* req; 

      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_REQUEST_FREE]); 
#endif   
      req = elg_request_get(*request);
      if (req && (req->type == SEND_PERS_REQ || req->type == RECV_PERS_REQ))
	elg_request_free(req);
    
      result = PMPI_Request_free(request);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Request_free(request);
    }
  
  return result;
}

/* -- MPI_Cancel -- */

int MPI_Cancel( MPI_Request* request )
{
  int result;

  if (IS_TRACE_ON)
    {
      struct ElgRequest* req; 

      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_CANCEL]); 
#endif   
      req = elg_request_get(*request);
      if (req && (req->type == SEND_PERS_REQ || req->type == RECV_PERS_REQ))
	elg_request_free(req);
    
      result = PMPI_Cancel(request);

#ifndef ELG_CSITE_INST 
      elg_exit(); 
#endif
      TRACE_ON();
    }
  else
    {
      result = PMPI_Cancel(request);
    }
  
  return result;
}

/* 
 *-----------------------------------------------------------------------------
 *
 * Collective communication
 *
 *-----------------------------------------------------------------------------
 */

/* -- MPI_Allreduce -- */

int MPI_Allreduce ( void* sendbuf, 
		    void* recvbuf, 
		    int count, 
		    MPI_Datatype datatype, 
		    MPI_Op op, 
		    MPI_Comm comm )
{
  int result, sz, N;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_ALLREDUCE]); 
#endif
      result = PMPI_Allreduce(sendbuf, recvbuf, count, datatype, op, comm);
      
      PMPI_Type_size(datatype, &sz);
      PMPI_Comm_size(comm, &N);
      elg_mpi_collexit(ELG_NO_ID, ELG_COMM_ID(comm), N*count*sz, count*sz);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Allreduce(sendbuf, recvbuf, count, datatype, op, comm);      
    }

  return result;
}

/* -- MPI_Barrier -- */

int MPI_Barrier( MPI_Comm comm )
{
  int result;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_BARRIER]); 
#endif
      result = PMPI_Barrier(comm);
      
      elg_mpi_collexit(ELG_NO_ID, ELG_COMM_ID(comm), 0, 0);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Barrier(comm);
    }

  return result;
}

/* -- MPI_Bcast -- */

int MPI_Bcast( void* buf, 
	       int count, 
	       MPI_Datatype datatype, 
	       int root, 
	       MPI_Comm comm )
{
  int result, sz, N, me;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_BCAST]); 
#endif
      result = PMPI_Bcast(buf, count, datatype, root, comm);
      
      PMPI_Type_size(datatype, &sz);
      PMPI_Comm_rank(comm, &me);
      if ( me == root )
        PMPI_Comm_size(comm, &N);
      else
        N = 0;
      elg_mpi_collexit(ELG_RANK_TO_PE(root, comm), ELG_COMM_ID(comm),
         	       N*count*sz, count*sz);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Bcast(buf, count, datatype, root, comm);
    }

  return result;
}

/* -- MPI_Gather -- */

int MPI_Gather(void* sendbuf,
	       int sendcount,
	       MPI_Datatype sendtype,
	       void* recvbuf,
	       int recvcount,
	       MPI_Datatype recvtype,
	       int root, 
	       MPI_Comm comm )
{
  int result, ssz, rsz, N, me;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_GATHER]); 
#endif
      result = PMPI_Gather(sendbuf, sendcount, sendtype,
			   recvbuf, recvcount, recvtype,
			   root, comm);
      
      PMPI_Type_size(sendtype, &ssz);
      PMPI_Comm_rank(comm, &me);
      if ( me == root ) {
        PMPI_Comm_size(comm, &N);
        PMPI_Type_size(recvtype, &rsz);
      } else {
        N = rsz = 0;
      }
      elg_mpi_collexit(ELG_RANK_TO_PE(root, comm), ELG_COMM_ID(comm),
		       sendcount*ssz, N*recvcount*rsz);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Gather(sendbuf, sendcount, sendtype,
			   recvbuf, recvcount, recvtype,
			   root, comm);
    }

  return result;
}

/* -- MPI_Reduce -- */

int MPI_Reduce( void* sendbuf, 
		void* recvbuf, 
		int count, 
		MPI_Datatype datatype, 
		MPI_Op op, 
		int root, 
		MPI_Comm comm )
{
  int result, sz, isroot, me;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_REDUCE]); 
#endif
      result = PMPI_Reduce(sendbuf, recvbuf, count, datatype, op, root, comm);
      
      PMPI_Type_size(datatype, &sz);
      PMPI_Comm_rank(comm, &me);
      isroot = ( me == root );
      elg_mpi_collexit(ELG_RANK_TO_PE(root, comm), ELG_COMM_ID(comm),
		       count*sz, isroot*count*sz);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Reduce(sendbuf, recvbuf, count, datatype, op, root, comm);
    }

  return result;
}

/* -- MPI_Gatherv -- */

int MPI_Gatherv( void* sendbuf,
		 int sendcount,
		 MPI_Datatype sendtype,
		 void* recvbuf, 
		 int *recvcounts,
		 int *displs,
		 MPI_Datatype recvtype, 
		 int root, 
		 MPI_Comm comm )
{
  int result, recvsz, sendsz, recvcount, me, N, i;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST 
      elg_enter(elg_mpi_regid[ELG__MPI_GATHERV]); 
#endif
      result = PMPI_Gatherv(sendbuf, sendcount, sendtype,
			    recvbuf, recvcounts, displs, recvtype, 
			    root, comm);

      PMPI_Type_size(recvtype, &recvsz);
      PMPI_Type_size(sendtype, &sendsz);
      PMPI_Comm_rank(comm, &me);

      recvcount = recvsz = 0;
      if ( me == root ) {
        PMPI_Comm_size(comm, &N);
        PMPI_Type_size(recvtype, &recvsz);
        for(i = 0; i<N; i++) recvcount += recvcounts[i];
      }

      elg_mpi_collexit(ELG_RANK_TO_PE(root, comm), ELG_COMM_ID(comm),
		       sendcount * sendsz, recvcount * recvsz);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Gatherv(sendbuf, sendcount, sendtype,
			    recvbuf, recvcounts, displs, recvtype, 
			    root, comm);
    }

  return result;
}

/* -- MPI_Allgather -- */

int MPI_Allgather( void* sendbuf,
		   int sendcount,
		   MPI_Datatype sendtype,
		   void* recvbuf, 
		   int recvcount,
		   MPI_Datatype recvtype, 
		   MPI_Comm comm )
{
  int result, recvsz, sendsz, N;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST  
      elg_enter(elg_mpi_regid[ELG__MPI_ALLGATHER]); 
#endif
      result = PMPI_Allgather(sendbuf, sendcount, sendtype,
			      recvbuf, recvcount, recvtype, 
			      comm);

      PMPI_Type_size(recvtype, &recvsz);
      PMPI_Type_size(sendtype, &sendsz);
      PMPI_Comm_size(comm, &N);

      elg_mpi_collexit(ELG_NO_ID, ELG_COMM_ID(comm),
		       N * sendcount * sendsz, N * recvcount * recvsz);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Allgather(sendbuf, sendcount, sendtype,
			      recvbuf, recvcount, recvtype, 
			      comm);
    }

  return result;
}

/* -- MPI_Allgatherv -- */

int MPI_Allgatherv( void* sendbuf,
                    int sendcount,
                    MPI_Datatype sendtype,
                    void* recvbuf,
                    int *recvcounts,
                    int *displs,
                    MPI_Datatype recvtype,
                    MPI_Comm comm )
{
  int result, recvcount, recvsz, sendsz, i, N;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST  
      elg_enter(elg_mpi_regid[ELG__MPI_ALLGATHERV]); 
#endif
      result = PMPI_Allgatherv(sendbuf, sendcount, sendtype,
			       recvbuf, recvcounts, displs, recvtype,
			       comm);


      PMPI_Type_size(recvtype, &recvsz);
      PMPI_Type_size(sendtype, &sendsz);
      PMPI_Comm_size(comm, &N);
      recvcount = 0;
      for(i = 0; i<N; i++) recvcount += recvcounts[i];

      elg_mpi_collexit(ELG_NO_ID, ELG_COMM_ID(comm),
		       N * sendcount * sendsz, recvcount * recvsz);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Allgatherv(sendbuf, sendcount, sendtype,
			       recvbuf, recvcounts, displs, recvtype,
			       comm);
    }

  return result;
}

/* -- MPI_Alltoall -- */

int MPI_Alltoall( void* sendbuf,
                  int sendcount,
                  MPI_Datatype sendtype,
                  void* recvbuf,
                  int recvcount,
                  MPI_Datatype recvtype,
                  MPI_Comm comm)
{
  int result, recvsz, sendsz, N;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST  
      elg_enter(elg_mpi_regid[ELG__MPI_ALLTOALL]); 
#endif
      result = PMPI_Alltoall(sendbuf, sendcount, sendtype,
			     recvbuf, recvcount, recvtype,
			     comm);

      PMPI_Type_size(recvtype, &recvsz);
      PMPI_Type_size(sendtype, &sendsz);
      PMPI_Comm_size(comm, &N);

      elg_mpi_collexit(ELG_NO_ID, ELG_COMM_ID(comm),
	               sendsz * sendcount * N, recvsz * recvcount * N);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Alltoall(sendbuf, sendcount, sendtype,
			     recvbuf, recvcount, recvtype,
			     comm);
    }

  return result;
}

/* -- MPI_Alltoallv -- */

int MPI_Alltoallv( void* sendbuf,
                   int *sendcounts,
                   int *sdispls,
                   MPI_Datatype sendtype,
                   void* recvbuf,
                   int *recvcounts,
                   int *rdispls,
                   MPI_Datatype recvtype,
                   MPI_Comm comm )
{
  int result, recvcount=0, sendcount=0, recvsz, sendsz, N, i;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST  
      elg_enter(elg_mpi_regid[ELG__MPI_ALLTOALLV]); 
#endif
      result = PMPI_Alltoallv(sendbuf, sendcounts,
                              sdispls, sendtype,
			      recvbuf, recvcounts,
                              rdispls, recvtype,
			      comm);
      PMPI_Type_size(recvtype, &recvsz);
      PMPI_Type_size(sendtype, &sendsz);
      PMPI_Comm_size(comm, &N);
      for(i = 0; i<N; i++)
        {
	  recvcount += recvcounts[i];
	  sendcount += sendcounts[i];
	}

      elg_mpi_collexit(ELG_NO_ID, ELG_COMM_ID(comm),
	               sendsz * sendcount, recvsz * recvcount);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Alltoallv(sendbuf, sendcounts,
                              sdispls, sendtype,
			      recvbuf, recvcounts,
                              rdispls, recvtype,
			      comm);
    }

  return result;
}

/* -- MPI_Scan -- */

int MPI_Scan( void* sendbuf,
	      void* recvbuf,
              int count,
              MPI_Datatype datatype,
              MPI_Op op,
              MPI_Comm comm  )
{
  int result, size, me, N;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST  
      elg_enter(elg_mpi_regid[ELG__MPI_SCAN]); 
#endif
      result = PMPI_Scan( sendbuf, recvbuf, count, datatype, op, comm );
      PMPI_Type_size(datatype, &size);
      PMPI_Comm_rank(comm, &me);
      PMPI_Comm_size(comm, &N);
      elg_mpi_collexit(ELG_NO_ID, ELG_COMM_ID(comm),
	               (N-me) * size * count, size * count);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Scan( sendbuf, recvbuf, count, datatype, op, comm );
    }

  return result;

}

/* -- MPI_Scatter -- */

int MPI_Scatter( void* sendbuf,
                 int sendcount,
                 MPI_Datatype sendtype,
                 void* recvbuf,
                 int recvcount,
                 MPI_Datatype recvtype,
                 int root,
                 MPI_Comm comm )
{
  int result, sendsz, recvsz, N, me;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST  
      elg_enter(elg_mpi_regid[ELG__MPI_SCATTER]); 
#endif
      result = PMPI_Scatter(sendbuf, sendcount, sendtype,
			   recvbuf, recvcount, recvtype,
			   root, comm);
      PMPI_Type_size(recvtype, &recvsz);
      PMPI_Comm_rank(comm, &me);
      if ( me == root ) {
        PMPI_Comm_size(comm, &N);
        PMPI_Type_size(sendtype, &sendsz);
      } else {
        N = sendsz = 0;
      }
      elg_mpi_collexit(ELG_RANK_TO_PE(root, comm), ELG_COMM_ID(comm),
		       N * sendcount * sendsz, recvcount * recvsz);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Scatter(sendbuf, sendcount, sendtype,
			   recvbuf, recvcount, recvtype,
			   root, comm);
    }

  return result;
}


/* -- MPI_Scatterv -- */

int MPI_Scatterv( void* sendbuf,
                  int *sendcounts,
                  int *displs,
                  MPI_Datatype sendtype,
                  void* recvbuf,
                  int recvcount,
                  MPI_Datatype recvtype,
                  int root,
                  MPI_Comm comm )
{
  int result, sendcount, recvsz, sendsz, me, N, i;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST  
      elg_enter(elg_mpi_regid[ELG__MPI_SCATTERV]); 
#endif
      result = PMPI_Scatterv(sendbuf, sendcounts, displs, sendtype,
			     recvbuf, recvcount, recvtype,
			     root, comm);
      sendcount = sendsz = 0;
      PMPI_Type_size(recvtype, &recvsz);
      PMPI_Comm_rank(comm, &me);
      if ( me == root ) {
        PMPI_Comm_size(comm, &N);
        PMPI_Type_size(sendtype, &sendsz);
        for(i = 0; i<N; i++) sendcount += sendcounts[i];
      }
      elg_mpi_collexit(ELG_RANK_TO_PE(root, comm), ELG_COMM_ID(comm),
		       sendcount * sendsz, recvcount * recvsz);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Scatterv(sendbuf, sendcounts, displs, sendtype,
			     recvbuf, recvcount, recvtype,
			     root, comm);
    }

  return result;
}

/* -- MPI_Reduce_scatter -- */

int MPI_Reduce_scatter( void* sendbuf,
			void* recvbuf,
			int *recvcounts,
			MPI_Datatype datatype,
			MPI_Op op,
			MPI_Comm comm )
{
  int result, i, size, N, count = 0;

  if (IS_TRACE_ON)
    {
      TRACE_OFF();
#ifndef ELG_CSITE_INST  
      elg_enter(elg_mpi_regid[ELG__MPI_REDUCE_SCATTER]); 
#endif
      result = PMPI_Reduce_scatter(sendbuf, recvbuf, recvcounts,
	                           datatype, op, comm);
      PMPI_Type_size(datatype, &size);
      PMPI_Comm_size(comm, &N);
      for(i = 0; i<N; i++) count += recvcounts[i];
      elg_mpi_collexit(ELG_NO_ID, ELG_COMM_ID(comm), count*size, count*size);
      TRACE_ON();
    }
  else
    {
      result = PMPI_Reduce_scatter(sendbuf, recvbuf, recvcounts,
	                           datatype, op, comm);
    }

  return result;
}
