/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee                                                **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "elg_rw.h"
#include "elg_error.h"

#include <stdio.h>
#include <string.h>

/* 
   This program is supposed to transform trace files compliant with
   older versions of EPILOG to the latest EPILOG version.

   However, since the program can not generate information that was
   not stored in the original file, its main purpose is to prevent
   tools expecting the latest EPILOG format from crashing. For
   example, the ELG_NODE record has a new field 'cr' (i.e., clock
   rate). elg_update simply extends the record to include this field
   and sets it to 1.0, since it doesn't know the clock rate of the
   target platform. 

   The following list explains the transformations performed by
   elg_update when upgrading to a specific version of the format. If
   you want to know all transformations performed on your file follow
   all the steps from the old version to the new version through all
   intermediate versions.

   1.0 -> 1.1
   ----------

   - Extension of the ELG_NODE record to include 'cr' field. The
     field is set to 1.0.

   - Insertion of a ELG_NUM_EVENTS record after counting all event
     records.

   - NOTE: There are no transformations of any hardware-counter
     related records, since the first version of the EPILOG library
     capable of writing this information was 1.1.
*/



int main(int argc, char **argv) 
{
  ElgIn*  in;
  ElgOut* out;
  ElgRec* rec;
  char*   infile_name;
  char*   outfile_name;
  long  eventc = 0;         /* event counter */

  if (argc != 3)
    elg_error_msg("Bad arguments\nUsage: elg_update <infile> <outfile>");
    
  infile_name  = argv[1]; 
  outfile_name = argv[2];

  if ( strcmp(infile_name, outfile_name) == 0 )
    elg_error_msg("<infile> has same name as <outfile> has");
    
  if ( (in = ElgIn_open(infile_name)) == NULL )
    elg_error_msg("Cannot open file %s\n", argv[1]);
  
  if ( ElgIn_get_version(in) == 1.1 )
    elg_error_msg("%s is already compliant with version 1.1\n", argv[1]);
  
  /* count number of event records */
  while((rec = ElgIn_read_record(in)))
    {
      if (ElgRec_is_event(rec))
	eventc++;
      ElgRec_free(rec);
    }

  /* reset read pointer to the beginning of the file */
  ElgIn_rewind(in);
  
  if ( (out = ElgOut_open(outfile_name)) == NULL )
    elg_error_msg("Cannot open file %s\n", outfile_name);
    
  
  while( (rec = ElgIn_read_record(in)) != NULL )
    {
      elg_ui1 type = ElgRec_get_type(rec);
      
      switch(type)
	{
	case ELG_NODE:
	  {
	    elg_ui4 nid  = ElgRec_read_ui4(rec);
	    elg_ui4 mid  = ElgRec_read_ui4(rec);
	    elg_ui4 cpuc = ElgRec_read_ui4(rec);
	    elg_ui4 nnid = ElgRec_read_ui4(rec); 

	    ElgRec* new_rec = ElgRec_create_NODE(nid, mid, cpuc, nnid, 1.0);
	    ElgRec_free(rec);
	    rec = new_rec;
	    break;
	  }
	case ELG_LAST_DEF:
	  {
	    ElgRec* new_rec = ElgRec_create_NUM_EVENTS(eventc);
	    ElgOut_write_record(out, new_rec);
	    ElgRec_free(new_rec);
	    break;
	  }
	}
      ElgOut_write_record(out, rec);
      ElgRec_free(rec);
    }
  ElgIn_close(in);
  ElgOut_close(out);
  
  return 0;
}


















