/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (C) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "Analyzer.h"

#include <cstdlib>
#include <iostream>

#include "CallTree.h"
#include "Pattern.h"
#include "PatternTree.h"
#include "Profile.h"
#include "Report.h"

using namespace earl;
using namespace std;


Analyzer::Analyzer(EventTrace* trace) : trace(trace)
{
  ptree = new PatternTree(trace, *this);
  profile = new Profile(trace, *this);
}

Analyzer::~Analyzer() 
{
  delete ptree;
  delete profile;
}

void
Analyzer::subscribe_event(etype type, Callback* cbobj)
{
  eventcbm[type].push_back(cbobj);
  if ( type == EXIT )
    {
      eventcbm[MPICEXIT].push_back(cbobj);
      eventcbm[OMPCEXIT].push_back(cbobj);
    }
}

void 
Analyzer::subscribe_cmpnd(cetype type, Callback* cbobj)
{
  ceventcbm[type].push_back(cbobj);
}
  
void 
Analyzer::run()
{
  int new_percent, percent = 0;

  for ( long pos = 1; pos <= trace->get_nevents(); pos++ )
    {
      this->step(pos);
      new_percent = ( 100 * pos ) / trace->get_nevents();
      if ( new_percent > percent )
	{
	  percent = new_percent;
	  cerr <<  '\r' << percent << " %"; 
	}
    }
  cerr << endl;

  // check for missing EXIT events (i.e., whether stack empty)
  long last_pos = trace->get_nevents();
  vector<Event> outv;
  for ( long loc_id = 0; loc_id < trace->get_nlocs(); loc_id++ )
    {
      trace->stack(outv, last_pos, loc_id);
      if ( ! outv.empty() )
	{
	  cerr << "EXPERT: Error in analysis. Trace file incomplete because of missing EXIT events." << endl;
	  exit(EXIT_FAILURE);
	}
    }

  CallTree ctree(trace);

  // postprocess profile
  profile->exclusify(ctree);
  
  // end() callbacks
  vector<long> keyv;
  ptree->get_keyv(keyv);
  for ( long i = 0; i < keyv.size(); i++ )
    ptree->get_data(keyv[i])->end(*profile, ctree);

  // generate report
  string cube_name = trace->get_name();
  cube_name.resize(cube_name.size() + 1);
  cube_name.replace(cube_name.size() - 4, 4, "cube");
  Report(*ptree, ctree, *trace).write(cube_name);

  return;
}

void 
Analyzer::step(long pos)
{
  Event curr = trace->event(pos);
  
  vector<cmp_event> ceventv;

  // callbacks for primitive events
  for ( long i = 0; i < eventcbm[curr.get_type()].size(); i++ )
    {
      cmp_event* ce;

      switch( curr.get_type() )
	{
	  case ALOCK:
	    ce = eventcbm[curr.get_type()][i]->alock(curr);
	    break;
	  case ENTER:
	    ce = eventcbm[curr.get_type()][i]->enter(curr);
	    break;
	  case EXIT:
	    ce = eventcbm[curr.get_type()][i]->exit(curr);
	    break;
	  case FORK:
	    ce = eventcbm[curr.get_type()][i]->fork(curr);
	    break;
	  case JOIN:
	    ce = eventcbm[curr.get_type()][i]->join(curr);
	    break;
	  case MPICEXIT:
	    ce = eventcbm[curr.get_type()][i]->mpicexit(curr);
	    if ( ce != NULL ) cequeue.push(ce);
	    ce = eventcbm[curr.get_type()][i]->exit(curr);
	    break;
	  case OMPCEXIT:
	    ce = eventcbm[curr.get_type()][i]->ompcexit(curr);
	    if ( ce != NULL ) cequeue.push(ce);
	    ce = eventcbm[curr.get_type()][i]->exit(curr);
	    break;
	  case RECV:
	    ce = eventcbm[curr.get_type()][i]->recv(curr);
	    break;
	  case RLOCK:
	    ce = eventcbm[curr.get_type()][i]->rlock(curr);
	    break;
	  case SEND:
	    ce = eventcbm[curr.get_type()][i]->send(curr);
	    break;
	}
      if ( ce != NULL )
	cequeue.push(ce);
    }
  
  // callbacks for compound events
  while ( cequeue.size() > 0 )
    {
      cmp_event* ce = cequeue.front();
      for ( long i = 0; i < ceventcbm[ce->type].size(); i++ )
	{
	  cmp_event* new_ce = ceventcbm[ce->type][i]->cmpnd(ce);
	  if ( new_ce != NULL )
	    cequeue.push(new_ce);
	}
      delete ce;
      cequeue.pop();
    }
}
